import requests
from lxml.etree import fromstring, XMLSyntaxError, _Element


class Finder:

    def _request(self, address, xpath):
        resp = requests.post(f"{self.srv_addr}/{address}", json={
            "session": self.session,
            "tab": self.tab_id,
            "xpath": xpath
        })
        data = resp.json()
        assert self.session == data.get('session')
        assert self.tab_id == data.get('tab')
        return data

    def _xpath_resolver(self, xpath):
        self_xpath = self.__dict__.get('xpath')
        if self_xpath:
            return f"{self_xpath}{xpath}"
        return xpath

    def find_element(self, xpath, remote=True):
        full_xpath = self._xpath_resolver(xpath)
        if remote:
            data = self._request('find_element', full_xpath)
            html = data.get('html')
        else:
            html = self.html.xpath(xpath)[0]
        assert html
        return WebElement(
            srv_addr=self.srv_addr,
            session=self.session,
            tab_id=self.tab_id,
            xpath=full_xpath,
            html=html
        )

    def find_elements(self, xpath, remote=True):
        full_xpath = self._xpath_resolver(xpath)
        if remote:
            data = self._request('find_elements', full_xpath)
            elements = data.get('elements')
        else:
            elements = self.html.xpath(xpath)
        assert type(elements) == list
        return [WebElement(
            srv_addr=self.srv_addr,
            session=self.session,
            tab_id=self.tab_id,
            xpath=full_xpath,
            html=element
        ) for element in elements]


class WebElement(Finder):
    @staticmethod
    def _parse(html):
        if type(html) == str:
            try:
                return fromstring(html)
            except XMLSyntaxError as e:
                print(e)
        elif isinstance(html, _Element):
            return html
        else:
            raise ValueError("Invalid HTML")

    def __init__(self, srv_addr, session, tab_id, xpath, html):
        self.srv_addr = srv_addr
        self.session = session
        self.tab_id = tab_id
        self.xpath = xpath
        self.html = self._parse(html)

    def get_attribute(self, attribute_name):
        return self.html.get(attribute_name)

    @property
    def text(self):
        return self.html.text

    @property
    def screenshot(self):
        resp = requests.post(f"{self.srv_addr}/tab/element/screenshot", json={
            "session": self.session,
            "tab": self.tab_id,
            "xpath": self.xpath,
        })
        data = resp.json()
        assert self.session == data.get('session')
        assert self.tab_id == data.get('tab')
        return data.get('screenshot')

    def click(self):
        pass

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.html.tag}>"



import requests
from parse_schema import Schema
from chrome_boss.element import Finder


class TreeChangedException(Exception):
    pass


class Cacher:

    def _update_elements(self, xpath, result):
        elements = self.html.xpath(xpath)
        parents = [elem.getparent() for elem in elements]
        if len(parents) > 1:
            if len(parents) == len(result):
                [elem.getparent().remove(elem) for elem in elements]
                [parent.insert(0, elem.html) for elem, parent in list(zip(result, parents))]
            else:
                TreeChangedException(f"Parents count changed from {len(parents)} to {len(result)}")
        else:
            print('one parent')

    def _update_element(self, xpath, result):
        if len(xpath._parts) == 1:
            self.html = result.html
        else:
            print('change tree')


class ChromeTab(Finder, Cacher):
    def __init__(self, srv_addr, tab_id, url, title, session):
        self.srv_addr = srv_addr
        self.tab_id = tab_id
        self.url = url
        self.title = title
        self.session = session
        self.html = None

    def get(self, url):
        resp = requests.post(f"{self.srv_addr}/tab/get", json={
            "session": self.session,
            "tab": self.tab_id,
            "url": url
        })
        if resp.status_code == 200:
            data = resp.json()
            self.url = url
            self.title = data.get('title')
        else:
            raise ValueError(f"status code: {resp.status_code}")

    def find_element(self, xpath, remote=True):
        result = super().find_element(xpath, remote=remote)
        if remote and isinstance(xpath, Schema):
            self._update_element(xpath, result)
        return result

    def find_elements(self, xpath, remote=True):
        result = super().find_elements(xpath, remote=remote)
        if remote and isinstance(xpath, Schema):
            self._update_elements(xpath, result)
        return result

    @property
    def screenshot(self):
        resp = requests.post(f"{self.srv_addr}/tab/screenshot", json={
            "session": self.session,
            "tab": self.tab_id,
        })
        data = resp.json()
        assert self.session == data.get('session')
        assert self.url == data.get('url')
        assert self.title == data.get('title')
        assert self.tab_id == data.get('tab')
        return data.get('screenshot')

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"<Tab: {self.title}>"



from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = ["requests", "parse_schema"]

setup(
    name="chrome_boss",
    version="0.5.0",
    author="zavx0z",
    author_email="zavx0z@yahoo.com",
    description="Chrome client",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/zavx0z/chrome_boss",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)



import uuid
import requests
from chrome_boss.tab import ChromeTab

__all__ = ['ChromeBoss']


class ChromeBoss:
    tabs = []

    def __init__(self, host='localhost', port=4444):
        self.url = f"http://{host}:{port}"
        resp = requests.post(f"{self.url}/init", json={"client_id": uuid.uuid4().hex})
        data = resp.json()
        self.session = data.get('session')
        self.tabs = [ChromeTab(
            srv_addr=self.url,
            tab_id=tab.get('id'),
            url=tab.get('url'),
            title=tab.get('title'),
            session=self.session,
        ) for tab in data.get('tabs')]
        self.session = resp.json().get('window_id')

    def tab(self, number):
        return self.tabs[number - 1]

    def destroy(self):
        print(f"Destroy {self.session}")

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.session}>"
